using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq; 
using UnityEngine.UI;
using System.IO;
using System.Globalization;
using Battlehub.RTGizmos;
using Battlehub.RTHandles;


namespace Doofenschmalphys
{
    public class MeasurementTool : MVRScript
    {
    

	

		
        JSONStorableString qstring;
        UIDynamicTextField stext;



        List<Slicer> allSlicers;
        List<RaySector> allSectors;
       public JSONStorableBool DeleteMode;
       public JSONStorableBool FreehandDraw;
       public JSONStorableBool UseInches;

       public JSONStorableFloat slicerDelta;

       
	

		public override void Init() {
			try {			

				//leftMultiplier = new JSONStorableFloat("Left Axis Offset", 0.0f, -50.0f, 50.0f, true, true);

            
                slicerDelta = new JSONStorableFloat("Slicer Accuracy (In Degrees)",1f,1f,360f,true,true);

                CreateSlider(slicerDelta,true);
	

                var spacer = CreateSpacer();
                spacer.height = 10;

				var btn2 = CreateButton("Delete Lines",true);
                btn2.button.onClick.AddListener(() =>
                {
               dan.Clear();
               qan.Clear();
                });

                


                var btn4 = CreateButton("Delete All Slicers",true);
                btn4.button.onClick.AddListener(() =>
                {
                    CSlicers();
                });


                

           

                qstring = new JSONStorableString("Chosen", "");
				stext = CreateTextField(qstring,false);
				stext.height=500.0f;
				qstring.val="";


             
                allSlicers = new List<Slicer>();
                allSectors = new List<RaySector>();
                 var btn7 = CreateButton("Add Slicer",true);
                btn7.button.onClick.AddListener(() =>
                {
                var q = new Slicer("");
                q.ForceSize(1.0f,0.01f,1.0f);
                allSlicers.Add(q);

               
                });



                 var btn8 = CreateButton("Toggle Handle Mode",true);
                btn8.button.onClick.AddListener(() =>
                {
                SmartToggleAll();
                });


                 var btn9 = CreateButton("Make Slicers",true);
                btn9.button.onClick.AddListener(() =>
                {
                GenerateSlicers();
                });


           
                
                DeleteMode = new JSONStorableBool("Delete Mode",false);
                var dmode = CreateToggle(DeleteMode,false);
                FreehandDraw = new JSONStorableBool("Enable Freehand Drawing",false);
                var drmode = CreateToggle(FreehandDraw,false);
                UseInches = new JSONStorableBool("Use Inches",false);
                var inchmode = CreateToggle(UseInches,false);



            JSONStorableString pluginVersionJSON = new JSONStorableString("PluginVersion","");
			UIDynamicTextField dtext = CreateTextField(pluginVersionJSON,false);
            pluginVersionJSON.val = @"Morph Analysis Tool v1.0 by Doofenschmalphys

Instructions:
The SPACEBAR can be used to:
    Delete Slicers
    Actually Do The Measuring
    Draw Freehand
Use the relevant checkboxes accordingly to do this.

To switch the handles for the Slicers between
(Move and Rotate) and (Resize), use the
Toggle Handle Mode button.

The Make Slicers button makes a bunch of preset Slicers
according to standardized body measurement systems.";
                
             
            }
			catch (Exception e) {
				SuperController.LogError("Exception caught: " + e);
			}
		}
		


public string numformat(float in2){
    in2 *= 200f;
    string in3 = " cm";
    if(UseInches.val) 
    {
        in2 *= 0.393701f;
        in3 = " in";
    }

    return in2.ToString() + in3;

}




 





        //This will be used more in later versions of this plugin.
        //It's like a "pie slice" raycast.
        //Given ray A and B, a RaySector will create N rays that are interpolations between A and B and return the results of the raycasts.
        public class RaySector
        {
            public Ray start;
            public Ray end;
            public float delta;
            public List<Vector3> points;
            

            public RaySector(Ray a, Ray b){
                start = a;
                end = b;
                delta = 0.1f;

            }



            public void PushDrawer(LineDrawer g){
                g.PushVertices(points,Color.red);
            }

            public void Execute(bool add){
                List<Vector3> vn = new List<Vector3>();
                for(float i = 0f;i<=1f;i+=delta){
                    Ray r = new Ray(Vector3.Lerp(start.origin,end.origin,i),Vector3.Lerp(start.direction,end.direction,i));
                    vn.Add(r.origin);
                    vn.Add(r.origin + (r.direction*3f));
                }

                if(add){

                foreach(var x in vn){
                    points.Add(x);
                }
                }
                if(!add) points = vn;

                
            }


            



                public void Execute(Func<Ray,Vector3> o,bool add){
                List<Vector3> vn = new List<Vector3>();
                for(float i = 0f;i<=1f;i+=delta){
                    Ray r = new Ray(Vector3.Lerp(start.origin,end.origin,i),Vector3.Lerp(start.direction,end.direction,i));
                

                    Vector3 hp = o(r);
                    if(hp!=Vector3.zero) { vn.Add(hp);}
                 

                }


            if(add){
                 foreach(var x in vn){
                    points.Add(x);
                }
            }
            if(!add) points = vn;

                
            }


        public float length(){
            float out2 = 0f;
            for(int i = 0; i < points.Count()-1;i++){
                out2 += Vector3.Distance(points[i],points[i+1]);
            }
            return out2;
        }



        public float RayTrain(List<Ray> train,LineDrawer g, Func<Ray,Vector3> a){
            float out2 = 0f;
            for(int i = 0; i < train.Count()-1;i+=1){
                start = train[i];
                end = train[i+1];
                Execute(a,false);
                out2+=length();
            }
            PushDrawer(g);
            return out2;
        }

        }




        //Wonderful magical raycast disks. 
        //Slicers execute a circular raycast, dividing 360 degrees into N rays and firing them inwards towards the center
        //Measuring the distances between these points is as close as you can get to a measuring tape.
        public class Slicer
        {
            public GameObject obj;
            public Vector3 size;
            public float delta;
            public bool HandleMode;


      



            public Slicer(String name)
            {

                //var c = new FreeControllerV3();
                //c.transform.position = Vector3.zero;

                //containingAtom.freeControllers
                obj = GameObject.CreatePrimitive(PrimitiveType.Sphere);
                obj.GetComponent<Renderer>().material.shader = Shader.Find("Transparent/Diffuse");
                obj.GetComponent<Renderer>().material.color = new Color(1f, 0.5f, 0.5f, 0.5f);
                //obj.GetComponent<Renderer>().enabled = false;
                obj.GetComponent<Collider>().enabled = false;

                delta = 1f;
                
                
                        
           
     
                HandleMode = true;
                obj.name = name;



            }



                  public bool CheckMouseDistance(){
                        Ray r =  SuperController.singleton.MonitorCenterCamera.ScreenPointToRay(Input.mousePosition);
                        Vector3 q =  SuperController.singleton.MonitorCenterCamera.transform.position - obj.transform.position;


            bool t = false;
            float j = Vector3.Angle(r.direction,q);
            //SuperController.LogMessage(Vector3.Angle(r.direction,q).ToString());
            if(j > 178f){
                t = true;
            }

            return t;
            }




            public void ForceSize(float x, float y, float z)
            {
                obj.transform.localScale = new Vector3(x, y, z);
                size = new Vector3(x,y,z);
            }



            public void ToggleSRP(){
               HandleMode = !HandleMode;
                UpHandles();
            }


            public void CullHandles(){

    
              
                Destroy(obj.GetComponent<ScaleHandle>());
                Destroy(obj.GetComponent<RotationHandle>());
                Destroy(obj.GetComponent<PositionHandle>());
            
            }

            public bool IsDragging(){
                 return obj.GetComponent<ScaleHandle>().IsDragging || obj.GetComponent<RotationHandle>().IsDragging || obj.GetComponent<PositionHandle>().IsDragging;

            }

            public void UpHandles(){
                if(HandleMode==true){
                    Destroy(obj.GetComponent<ScaleHandle>());
                    if(!obj.GetComponent<PositionHandle>()) obj.AddComponent<PositionHandle>();
                    if(!obj.GetComponent<RotationHandle>()) obj.AddComponent<RotationHandle>();
                }
                if(HandleMode==false){
                    Destroy(obj.GetComponent<RotationHandle>());
                    Destroy(obj.GetComponent<PositionHandle>());
                    if(!obj.GetComponent<ScaleHandle>())  obj.AddComponent<ScaleHandle>();
                }

            }


    

            public float ExecuteSlice(LineDrawer g,Func<Ray,Vector3> o){
                size = obj.transform.localScale;
                List<Vector3> vs = new List<Vector3>();
                Vector3 p1 = Vector3.zero;
                p1.x += 0.5f;

              
                for(float i = 0; i<360;i+=delta){
                    var q = new Vector3(0,i,0);
                    var p3 = Rotated(p1,q,Vector3.zero);
                    p3.z *= size.z;
                    p3.x *= size.x;
                    p3 += obj.transform.position;
                    p3 = Rotated(p3,obj.transform.rotation,obj.transform.position);


                      Vector3 j = o(new Ray(p3,obj.transform.position-p3));
                      bool truth = Vector3.Distance(j,obj.transform.position) < Mathf.Max(size.x,size.z);
                        if(j!=Vector3.zero && truth)
                            {
                            vs.Add(j);
                            }



                  

                }


                
                float cdist = 0f;
                for(int j = 0; j<vs.Count();j+=2){
                    cdist += Vector3.Distance(vs[j],vs[j+1]);
                }
                


            


            g.PushVertices(vs,Color.blue);
            return cdist;
            }
        }




         public class LineDrawer
    {
        private readonly Mesh _mesh;
        private readonly List<Vector3> _vertices;
        private readonly List<Color> _colors;


        public LineDrawer()
        {
            _mesh = new Mesh();
            _vertices = new List<Vector3>();
            _colors = new List<Color>();
        }


        public void ResolveLines(){
            var q = _vertices.Count();
            _vertices.Add(_vertices[_vertices.Count()-1]);

        }

        public void PushLine(Vector3 start, Vector3 stop, Color color)
        {
            _vertices.Add(start);
            _vertices.Add(stop);
            _colors.Add(color);
            _colors.Add(color);
        }

        public void PushVertices(List<Vector3> pts,Color color){
           foreach(Vector3 pt in pts){
               _vertices.Add(pt);
               _colors.Add(color);
           }
        }

        public void Draw(Material material, int layer = 0)
        {
            var idx = Enumerable.Range(0, _vertices.Count);

            _mesh.vertices = _vertices.ToArray();
            _mesh.colors = _colors.ToArray();
            _mesh.uv = idx.Select(_ => Vector2.zero).ToArray();
            _mesh.normals = idx.Select(_ => Vector3.zero).ToArray();
            _mesh.SetIndices(idx.ToArray(), MeshTopology.Lines, 0);
            _mesh.RecalculateBounds();

            Graphics.DrawMesh(_mesh, Matrix4x4.identity, material, layer);
        }
        

        public void Clear()
        {
            _vertices.Clear();
            _colors.Clear();
        }
    }




/*
public void GenerateSectors(){
var Colliders = containingAtom.GetComponentsInChildren<Collider>();
var rs = Colliders.First(w => w.name == "rArmControl");
var ls = Colliders.First(w => w.name == "lArmControl");  
var chest = Colliders.First(w => w.name == "chestControl"); 

var rt = Colliders.First(w => w.name == "rThighControl"); 
var rn = Colliders.First(w => w.name == "rKneeControl"); 
var rf = Colliders.First(w => w.name == "rFootControl"); 

var ad = (rs.transform.position-chest.transform.position) - chest.transform.forward;
var a = new Ray(rs.transform.position - chest.transform.forward, chest.transform.forward);
var bd = (ls.transform.position-chest.transform.position) - chest.transform.forward;
var b = new Ray(ls.transform.position - chest.transform.forward,chest.transform.forward);

var c = new RaySector(a,b);
c.Execute(BodyRay,false);


//c.Execute(dan);


c.start = new Ray(rt.transform.position + rt.transform.right,-rt.transform.right);
c.end = new Ray(rn.transform.position + rn.transform.right, -rn.transform.right);


float l = 0f;

c.Execute(BodyRay,false);

l+= c.length();

c.start = new Ray(rn.transform.position + rn.transform.right,-rn.transform.right);
var reald = rn.transform.position + (rn.transform.up * -Vector3.Distance(rf.transform.position,rn.transform.position));
c.end = new Ray(reald + rn.transform.right,-rn.transform.right);

//c.Execute(dan);
c.Execute(BodyRay,false);

l+= c.length();




qstring.val = (l*100).ToString();


}
*/


public void GenerateSlicers(){


var Colliders = containingAtom.GetComponentsInChildren<Collider>();
var neck = Colliders.First(w => w.name == "neckControl");    
    
var rthigh = Colliders.First(w => w.name == "rThighControl");
var lthigh = Colliders.First(w => w.name == "lThighControl");      
var rKnee = Colliders.First(w => w.name == "rKneeControl");    
var head = Colliders.First(w => w.name == "headControl");        
var waist = Colliders.First(w => w.name == "abdomen2Control");
var waist2 = Colliders.First(w => w.name == "abdomenControl");


var lnip = Colliders.First(w => w.name == "lNippleControl");
var rnip = Colliders.First(w => w.name == "rNippleControl");

var arm1= Colliders.First(w => w.name == "rArmControl");
var arm2 = Colliders.First(w => w.name == "rElbowControl");
var hand = Colliders.First(w => w.name == "rHandControl");
var chest = Colliders.First(w => w.name == "chestControl");


var qn = new Slicer("Neck");
qn.ForceSize(0.25f,0.01f,0.25f);
qn.obj.transform.position = Vector3.Lerp(head.transform.position,neck.transform.position,0.9f);
qn.obj.transform.rotation = neck.transform.rotation;



var headq = new Slicer("Forehead");
headq.ForceSize(0.25f,0.01f,0.25f);
headq.obj.transform.position = head.transform.position + head.transform.up * 0.05f;
headq.obj.transform.rotation = head.transform.rotation;


var thigh = new Slicer("Thigh");
thigh.ForceSize(0.25f,0.01f,0.25f);
thigh.obj.transform.position = rthigh.transform.position + rthigh.transform.up * -0.2f;
thigh.obj.transform.rotation = rthigh.transform.rotation;

var K = new Slicer("Knee");
K.ForceSize(rKnee.bounds.extents.x*2f,0.01f,rKnee.bounds.extents.z*2f);
K.obj.transform.position = rKnee.transform.position;
K.obj.transform.rotation = rKnee.transform.rotation;


var W = new Slicer("Waist");
W.ForceSize(waist.bounds.extents.x * 3f,0.01f,waist.bounds.extents.z * 3f);
W.obj.transform.position = waist.transform.position;
W.obj.transform.rotation = waist2.transform.rotation;









var NIPS = new Slicer("Bust");
NIPS.ForceSize(0.8f,0.01f,0.8f);
NIPS.obj.transform.position = Vector3.Lerp(lnip.transform.position,rnip.transform.position,0.5f);
NIPS.obj.transform.rotation = Quaternion.Lerp(chest.transform.rotation,neck.transform.rotation,0.5f);
NIPS.obj.transform.position += NIPS.obj.transform.forward * -0.2f/4;


var UNB = new Slicer("Under Bust");
UNB.ForceSize(0.8f,0.01f,0.8f);
UNB.obj.transform.position = Vector3.Lerp(NIPS.obj.transform.position,waist2.transform.position,0.2f);
UNB.obj.transform.rotation = NIPS.obj.transform.rotation;




var OVB = new Slicer("Over Bust");
OVB.ForceSize(0.5f,0.01f,0.5f);
OVB.obj.transform.position = Vector3.Lerp(NIPS.obj.transform.position,neck.transform.position,0.5f);
OVB.obj.transform.rotation = NIPS.obj.transform.rotation;


var ArmThick1 = new Slicer("Bicep Thickness");
ArmThick1.ForceSize(0.25f,0.01f,0.25f);
ArmThick1.obj.transform.position = Vector3.Lerp(arm1.transform.position,arm2.transform.position,0.5f);
ArmThick1.obj.transform.rotation = Quaternion.Lerp(arm1.transform.rotation,arm2.transform.rotation,0.5f);
ArmThick1.obj.transform.up = arm1.transform.right;

var ArmThick2 = new Slicer("Forearm Thickness");
ArmThick2.ForceSize(0.25f,0.01f,0.25f);
ArmThick2.obj.transform.position = Vector3.Lerp(arm2.transform.position,hand.transform.position,0.5f);
ArmThick2.obj.transform.rotation = Quaternion.Lerp(arm2.transform.rotation,hand.transform.rotation,0.5f);
ArmThick2.obj.transform.up = arm2.transform.right;

var hips = new Slicer("Hips");
float d = Vector3.Distance(rthigh.transform.position,lthigh.transform.position) + rthigh.bounds.extents.z*2;
hips.ForceSize(d,0.01f,d);
hips.obj.transform.position = Vector3.Lerp(rthigh.transform.position,lthigh.transform.position,0.5f);
hips.obj.transform.rotation = waist2.transform.rotation;



allSlicers.Add(qn);
allSlicers.Add(headq);
allSlicers.Add(thigh);
allSlicers.Add(K);
allSlicers.Add(W);
allSlicers.Add(NIPS);
allSlicers.Add(ArmThick1);
allSlicers.Add(ArmThick2);
allSlicers.Add(hips);
allSlicers.Add(OVB);
allSlicers.Add(UNB);

}











public void CSlicers(){
  while(allSlicers.Count() > 0){
      Slicer s = allSlicers[0];
      Destroy(s.obj);
      allSlicers.Remove(s);
  }
}





public void SliceAll()
{
    int count = 0;
    float runningDistance = 0f;
    float countedSlices = 0f;
    qstring.val = "";
    foreach (Slicer s in allSlicers)
    {
        s.delta = slicerDelta.val;
        var p = s.ExecuteSlice(dan, BodyRay);




        String name = s.obj.name;
        if (name == "") name = count.ToString();
        qstring.val += name + " " + numformat(p) + "\n";
        runningDistance += p;
        countedSlices += 1f;
        count++;
    }

    qstring.val += "Total: " + numformat(runningDistance) + "\n";
    qstring.val += "Average: " + numformat(runningDistance / countedSlices);


}


public void SmartToggleAll()
{
    foreach (Slicer s in allSlicers)
    {
        s.ToggleSRP();
    }

}





LineDrawer dan = new LineDrawer();
RaySector jan = new RaySector(new Ray(Vector3.zero,Vector3.zero),new Ray(Vector3.zero,Vector3.zero));
List<Ray> qan = new List<Ray>();

        //public DAZMesh megamesh;

protected void Freehand()
{
    qan.Add(SuperController.singleton.MonitorCenterCamera.ScreenPointToRay(Input.mousePosition));
    jan.delta = 1;
    var z = jan.RayTrain(qan, dan, BodyRay);
    qstring.val = numformat(z/2);

}
protected void Calibrate()
{
    SliceAll();
    //GenerateSectors();
}



public static Vector3 Rotated(Vector3 vector, Quaternion rotation, Vector3 pivot = default(Vector3)) {
    return rotation * (vector - pivot) + pivot;
}
 

 
public static Vector3 Rotated(Vector3 vector, Vector3 rotation, Vector3 pivot = default(Vector3)) {
    return Rotated(vector, Quaternion.Euler(rotation), pivot);
}
 
public static Vector3 Rotated(Vector3 vector, float x, float y, float z, Vector3 pivot = default(Vector3)) {
    return Rotated(vector, Quaternion.Euler(x, y, z), pivot);
}




















protected Vector3 BodyRay(Ray r){
    var z = containingAtom.GetComponentsInChildren<Collider>();
    RaycastHit hit = new RaycastHit();
    float md = 100f;
    Vector3 p= Vector3.zero;
    foreach(var collider in z){
    if(!collider.name.Contains("Control") && !collider.name.Contains("Link")){
               if(collider.Raycast(r, out hit, Mathf.Infinity)){
                   if(Vector3.Distance(hit.point,r.origin) < md){
                       md = Vector3.Distance(hit.point,r.origin);
                       p = hit.point;
                   }
               }


    }
    }
return p;
}









		





	

		

	



		// Start is called once before Update or FixedUpdate is called and after Init()
		void Start() {
			try {
	





    
			}
			catch (Exception e) {
				SuperController.LogError("Exception caught: " + e);
			}
		}


       
		// Update is called with each rendered frame by Unity
        Slicer ChosenOne;
		void Update() {
        
            //qstring.val += (100*runningDistance).ToString()  + "\n";
      

           foreach(var s in allSlicers){
             
               if(s.CheckMouseDistance()){
                s.UpHandles();
                ChosenOne = s;
              
               }else{

                if(s!=ChosenOne) s.CullHandles();
               }
               
                  //SuperController.LogMessage(s.IsDragging().ToString());
           }


		}

		// FixedUpdate is called with each physics simulation frame by Unity
		void FixedUpdate() {
            dan.Draw(new Material(Shader.Find("Custom/Subsurface/EmissiveGlow")),-3);
            //dan.Draw(new Material(Shader.Find("Custom/DebugTangents")),-3);
          
           
          
            if(Input.GetKey("space")){
            

            bool truth = true;

            if(DeleteMode.val){
            //SuperController.LogMessage("deleting");
            Destroy(ChosenOne.obj);
            allSlicers.Remove(ChosenOne);
            truth = false;
            }


            if(FreehandDraw.val){
                Freehand();
                truth = false;
            }



                
            if(truth){
                qstring.val = "TAKING MEASUREMENTS! Please wait~";
                Calibrate();
            }
               
            }

            //qstring.val = "";
        }



    void OnDestroy()
    {
        dan.Clear();
        qan.Clear();
        jan = null;
        CSlicers();

    }

	}
}
 